from typing import Any
from dataclasses import dataclass
from ..network import Network
from ..solution import NetworkSolution
from . import matrix_operations as mo
from . import label_mapping as map
from . import node_analysis as na

@dataclass(frozen=True)
class NodalAnalysisSolution:
    network: Network
    solution_vector: tuple
    label_mappings_factory: map.LabelMappingsFactory

    @property
    def label_mappings(self) -> map.NetworkLabelMappings:
        return self.label_mappings_factory(self.network)

    @property
    def _potentials(self) -> tuple:
        return tuple(self.solution_vector[i] for i in sorted(self.label_mappings.node_mapping.values))

    @property
    def _voltage_source_currents(self) -> tuple:
        all_indices = set(range(len(self.solution_vector)))
        remaining_indices = all_indices - set(self.label_mappings.node_mapping.values)
        return tuple(self.solution_vector[i] for i in sorted(remaining_indices))

    def _is_floating_node(self, node_id: str) -> bool:
        all_node_ids = {b.node1 for b in self.network.branches} | {b.node2 for b in self.network.branches}
        if node_id not in all_node_ids:
            raise KeyError(f"Node '{node_id}' not found in the network.")
        return node_id not in self.network.node_labels

    def _is_floating_branch(self, branch_id: str) -> bool:
        all_branch_ids = [b.id for b in self.network.branches]
        if branch_id not in all_branch_ids:
            raise KeyError(f"Branch '{branch_id}' not found in the network.")
        return branch_id not in self.network.branch_ids

    def get_potential(self, node_id: str) -> complex:
        if self._is_floating_node(node_id):
            return float('nan')
        if node_id == self.network.reference_node_label:
            return 0
        return self._potentials[self.label_mappings.node_mapping[node_id]]

    def get_current(self, branch_id: str) -> complex:
        if self._is_floating_branch(branch_id):
            return float('nan')
        if branch_id in self.label_mappings.voltage_source_mapping.keys:
            return self._voltage_source_currents[self.label_mappings.voltage_source_mapping[branch_id]]
        if self.network[branch_id].element.is_ideal_current_source:
            return complex(self.network[branch_id].element.I)
        if self.network[branch_id].element.is_current_source:
            return - (self.network[branch_id].element.I + self.get_voltage(branch_id)/self.network[branch_id].element.Z)
        branch = self.network[branch_id]
        return self.get_voltage(branch_id)/branch.element.Z

    def get_voltage(self, branch_id: str) -> Any:
        if self._is_floating_branch(branch_id):
            return float('nan')
        phi1 = self.get_potential(self.network[branch_id].node1)
        phi2 = self.get_potential(self.network[branch_id].node2)
        return phi1-phi2

    def get_power(self, branch_id: str) -> Any:
        return self.get_voltage(branch_id)*self.get_current(branch_id).conjugate()

def numeric_nodal_analysis_bias_point_solution(network: Network, label_mappings_factory: map.LabelMappingsFactory = map.default_label_mappings_factory) -> NetworkSolution:
        return NodalAnalysisSolution(
            network=network,
            solution_vector=na.nodal_analysis_solution(network, matrix_ops=mo.NumPyMatrixOperations(), label_mappings_factory=label_mappings_factory),
            label_mappings_factory=label_mappings_factory
        )

def symbolic_nodal_analysis_bias_point_solution(network: Network, label_mappings_factory: map.LabelMappingsFactory = map.default_label_mappings_factory) -> NetworkSolution:
        return NodalAnalysisSolution(
            network=network,
            solution_vector=na.nodal_analysis_solution(network, matrix_ops=mo.SymPyMatrixOperations(), label_mappings_factory=label_mappings_factory),
            label_mappings_factory=label_mappings_factory
        )


from typing import Callable
from dataclasses import dataclass
from ..network import Network

class DistinctValues(Exception):
    ...

@dataclass(frozen=True)
class LabelMapping:
    mapping: dict[str, int]

    def __post_init__(self) -> None:
        if len(set(self.mapping.values())) != len(self.mapping):
            raise DistinctValues

    def __getitem__(self, label: str) -> int:
        return self.mapping[label]

    def __call__(self, *labels: str) -> tuple[int, ...]:
        return tuple(self[label] for label in labels)

    def filter_keys(self, filter_fcn: Callable[[str], bool]) -> "LabelMapping":
        return LabelMapping({k: self.mapping[k] for k in self.mapping.keys() if filter_fcn(k)})

    @property
    def keys(self) -> list[str]:
        return list(self.mapping.keys())

    @property
    def values(self) -> list[int]:
        return list(self.mapping.values())

    @property
    def N(self) -> int:
        return len(self.mapping)

    def __iter__(self):
        return iter(self.mapping.keys())

def alphabetic_node_mapper(network: Network) -> LabelMapping:
    node_labels_without_zero = [label for label in sorted(network.node_labels) if label != network.reference_node_label] 
    return LabelMapping({k: v for v, k in enumerate(node_labels_without_zero)})

def alphabetic_source_mapper(network: Network) -> LabelMapping:
    current_source_labels = [b.id for b in network.branches if b.element.is_current_source]
    voltage_source_labels = [b.id for b in network.branches if b.element.is_ideal_voltage_source]
    sorted_soruce_labels = sorted(current_source_labels+voltage_source_labels)
    return LabelMapping({k: v for v, k in enumerate(sorted_soruce_labels)})

def alphabetic_current_source_mapper(network: Network) -> LabelMapping:
    current_source_labels = sorted([b.id for b in network.branches if b.element.is_current_source])
    return LabelMapping({k: v for v, k in enumerate(current_source_labels)})

def alphabetic_voltage_source_mapper(network: Network) -> LabelMapping:
    voltage_source_labels = sorted([b.id for b in network.branches if b.element.is_ideal_voltage_source])
    return LabelMapping({k: v for v, k in enumerate(voltage_source_labels)})

LabelMapper = Callable[[Network], LabelMapping]

@dataclass(frozen=True)
class NetworkLabelMappings:
    network: Network
    node_mapper: LabelMapper
    source_mapper: LabelMapper
    current_source_mapper: LabelMapper
    voltage_source_mapper: LabelMapper

    @property
    def node_mapping(self) -> LabelMapping:
        return self.node_mapper(self.network)

    @property
    def source_mapping(self) -> LabelMapping:
        return self.source_mapper(self.network)

    @property
    def current_source_mapping(self) -> LabelMapping:
        return self.current_source_mapper(self.network)

    @property
    def voltage_source_mapping(self) -> LabelMapping:
        return self.voltage_source_mapper(self.network)

LabelMappingsFactory = Callable[[Network], NetworkLabelMappings]

def default_label_mappings_factory(network: Network) -> NetworkLabelMappings:
    return NetworkLabelMappings(
        network=network,
        node_mapper=alphabetic_node_mapper,
        source_mapper = alphabetic_source_mapper,
        current_source_mapper=alphabetic_current_source_mapper,
        voltage_source_mapper=alphabetic_voltage_source_mapper
    )


from . import node_analysis as na
from .solution import numeric_nodal_analysis_bias_point_solution
from ..network import Network

def open_circuit_voltage(network: Network, node1: str, node2: str) -> complex:
    solution = numeric_nodal_analysis_bias_point_solution(network)
    if node1 == node2:
        return 0
    phi1 = solution.get_potential(node_id=node1)
    phi2 = solution.get_potential(node_id=node2)
    return phi1-phi2

def short_circuit_current(network: Network, node1: str, node2: str) -> complex:
    Z = complex(na.open_circuit_impedance(network, node1, node2))
    V = open_circuit_voltage(network, node1, node2)
    return V/Z



from .circuit import Circuit, transform, frequency_components, transform_symbolic_circuit
from ..SignalProcessing.types import TimeDomainFunction, FrequencyDomainSeries, TimeDomainSeries
from ..SignalProcessing.state_space_model import NumericStateSpaceModel, continuous_state_space_solver
from ..Network.NodalAnalysis.solution import numeric_nodal_analysis_bias_point_solution, symbolic_nodal_analysis_bias_point_solution
from .state_space_model import numeric_state_space_model_constructor, StateSpaceMatrixConstructor
from ..Network.solution import NetworkSolution, NetworkSolver
from typing import Any
from dataclasses import dataclass
from typing import Protocol
import numpy as np
import sympy as sp

class CircuitSolution(Protocol):
    def get_voltage(self, component_id: str) -> Any: ... 

    def get_current(self, component_id: str) -> Any: ...

    def get_potential(self, node_id: str) -> Any: ...

    def get_power(self, component_id: str) -> Any: ...

@dataclass(frozen=True)
class ScalarCircuitSolution:
    solution: NetworkSolution

@dataclass(frozen=True)
class VectorCircuitSolution:
    solutions: list[ScalarCircuitSolution]

@dataclass(frozen=True)
class EmptySolution:
    def get_voltage(self, component_id: str) -> Any:
        return None

    def get_current(self, component_id: str) -> Any:
        return None

    def get_potential(self, node_id: str) -> Any:
        return None

    def get_power(self, component_id: str) -> Any:
        return None

@dataclass(frozen=True)
class DCSolution(ScalarCircuitSolution):
    def get_voltage(self, component_id: str) -> float:
        return self.solution.get_voltage(component_id).real

    def get_current(self, component_id: str) -> float:
        return self.solution.get_current(component_id).real

    def get_potential(self, node_id: str) -> float:
        return self.solution.get_potential(node_id).real

    def get_power(self, component_id: str) -> float:
        return self.get_voltage(component_id)*self.get_current(component_id)

@dataclass(frozen=True)
class ComplexSolution(ScalarCircuitSolution):
    w: float = 0
    peak_values: bool = False

    def get_voltage(self, component_id: str) -> complex:
        return self.solution.get_voltage(component_id)

    def get_current(self, component_id: str) -> complex:
        return self.solution.get_current(component_id)

    def get_potential(self, node_id: str) -> complex:
        return self.solution.get_potential(node_id)

    def get_power(self, component_id: str) -> complex:
        if self.peak_values:
            return 1/2*self.get_voltage(component_id)*np.conj(self.get_current(component_id))
        return self.get_voltage(component_id)*np.conj(self.get_current(component_id))

class SymbolicSolution(ScalarCircuitSolution):
    def get_voltage(self, component_id: str) -> Any:
        return self.solution.get_voltage(component_id).simplify().nsimplify()

    def get_current(self, component_id: str) -> Any:
        return self.solution.get_current(component_id).simplify().nsimplify()

    def get_potential(self, node_id: str) -> Any:
        return self.solution.get_potential(node_id).simplify().nsimplify()

    def get_power(self, component_id: str) -> Any:
        return (self.get_voltage(component_id)*self.get_current(component_id)).simplify().nsimplify()

@dataclass(frozen=True)
class TimeDomainSolution(VectorCircuitSolution):
    solutions: list[ComplexSolution]
    w: list[float]

    def get_voltage(self, component_id: str) -> TimeDomainFunction:
        voltages = [solution.get_voltage(component_id) for solution in self.solutions]
        return np.vectorize(lambda t: np.array(np.sum([np.abs(V)*np.cos(w*t+np.angle(V)) for V, w in zip(voltages, self.w)])))

    def get_current(self, component_id: str) -> TimeDomainFunction:
        currents = [solution.get_current(component_id) for solution in self.solutions]
        return np.vectorize(lambda t: np.array(np.sum([np.abs(V)*np.cos(w*t+np.angle(V)) for V, w in zip(currents, self.w)])))

    def get_potential(self, node_id: str) -> TimeDomainFunction:
        potentials = [solution.get_potential(node_id) for solution in self.solutions]
        return np.vectorize(lambda t: np.array(np.sum([np.abs(phi)*np.cos(w*t+np.angle(phi)) for phi, w in zip(potentials, self.w)])))

    def get_power(self, component_id: str) -> TimeDomainFunction:
        voltage = self.get_voltage(component_id)
        current = self.get_current(component_id)
        return lambda t: np.array(voltage(t))*np.array(current(t))

@dataclass(frozen=True)
class FrequencyDomainSolution(VectorCircuitSolution):
    solutions: list[ComplexSolution]
    w: np.ndarray

    def get_voltage(self, component_id: str) -> FrequencyDomainSeries:
        voltages = 1/2*np.array([solution.get_voltage(component_id) for solution in self.solutions])
        return np.array(self.w), voltages

    def get_current(self, component_id: str) -> FrequencyDomainSeries:
        currents = 1/2*np.array([solution.get_current(component_id) for solution in self.solutions])
        return np.array(self.w), currents

    def get_potential(self, node_id: str) -> FrequencyDomainSeries:
        potentials = 1/2*np.array([solution.get_potential(node_id) for solution in self.solutions])
        return np.array(self.w), potentials

    def get_power(self, component_id: str) -> FrequencyDomainSeries:
        power = np.array([solution.get_power(component_id) for solution in self.solutions])
        return np.array(self.w), power

@dataclass(frozen=True)
class TransientSolution:
    t: np.ndarray
    ssm: StateSpaceMatrixConstructor
    u: np.ndarray
    x: np.ndarray

    def get_voltage(self, component_id: str) -> TimeDomainSeries:
        c, d = self.ssm.c_d_row_for_voltage(component_id)
        return self.t, np.reshape(c@self.x + d@self.u, (-1,))

    def get_current(self, component_id: str) -> TimeDomainSeries:
        c, d = self.ssm.c_d_row_for_current(component_id)
        return self.t, np.reshape(c@self.x + d@self.u, (-1,))

    def get_potential(self, node_id: str) -> TimeDomainSeries:
        c, d = self.ssm.c_d_row_for_potential(node_id)
        return self.t, np.reshape(c@self.x + d@self.u, (-1,))

    def get_power(self, component_id: str) -> TimeDomainSeries:
        return self.t, self.get_voltage(component_id)[1]*self.get_current(component_id)[1]

def dc_solution(circuit: Circuit, solver: NetworkSolver = numeric_nodal_analysis_bias_point_solution) -> DCSolution:
    network = transform(circuit, w=[0])[0]
    solution = solver(network)
    return DCSolution(solution=solution)

def complex_solution(circuit: Circuit, w: float = 0, peak_values: bool = False, solver: NetworkSolver = numeric_nodal_analysis_bias_point_solution) -> ComplexSolution:
    network = transform(circuit, w=[w], rms=not peak_values)[0]
    solution = solver(network)
    return ComplexSolution(solution=solution, w=w, peak_values=peak_values)

def symbolic_solution(circuit: Circuit, s: sp.core.symbol.Symbol = sp.Symbol('s', complex=True), solver: NetworkSolver = symbolic_nodal_analysis_bias_point_solution) -> SymbolicSolution:
    network = transform_symbolic_circuit(circuit, s=s)
    solution = solver(network)
    return SymbolicSolution(solution=solution)

def time_domain_solution(circuit: Circuit, w_max: float = 0, solver: NetworkSolver = numeric_nodal_analysis_bias_point_solution) -> TimeDomainSolution:
    w = frequency_components(circuit, w_max)
    solutions = [complex_solution(circuit, w=w_, peak_values=True, solver=solver) for w_ in w]
    return TimeDomainSolution(solutions=solutions, w=w)

def frequency_domain_solution(circuit: Circuit, w_max: float = 0, solver: NetworkSolver = numeric_nodal_analysis_bias_point_solution) -> FrequencyDomainSolution:
    w = np.array(frequency_components(circuit, w_max))
    solutions = [complex_solution(circuit, w=w_, peak_values=False, solver=solver) for w_ in w]
    return FrequencyDomainSolution(solutions=solutions, w=w)

def transient_solution(circuit: Circuit, tin: np.ndarray = np.zeros(0), input: dict[str, TimeDomainFunction] = {'': lambda t: np.zeros(0)}) -> TransientSolution:
    def _input_fcn(input_id: str) -> TimeDomainFunction:
        try:
            return input[input_id]
        except KeyError as e:
            raise KeyError(f'Input element with id "{input_id}" not defined.') from e
    ssm = numeric_state_space_model_constructor(circuit)
    u = np.array([_input_fcn(input_id)(tin) for input_id in ssm.sources])
    tout, x, _ = continuous_state_space_solver(
        NumericStateSpaceModel(A=ssm.A, B=ssm.B, C=np.eye(ssm.A.shape[0]), D=np.zeros((ssm.A.shape[0], ssm.B.shape[1]))),
        u.T,
        tin,
        np.zeros((ssm.A.shape[0], ))
    )
    x = np.reshape(x, (x.shape[0], ssm.A.shape[0])).T
    return TransientSolution(t=tout, ssm=ssm, u=u, x=x)
        
        


from . import elements as elm
from dataclasses import dataclass

class AmbiguousBranchIDs(Exception): pass

@dataclass(frozen=True)
class Branch:
    node1 : str
    node2 : str
    element : elm.NortenTheveninElement

    @property
    def id(self) -> str:
        return self.element.name

@dataclass(frozen=True)
class Network:
    branches: list[Branch]
    reference_node_label: str = '0'

    def __post_init__(self):
        branch_ids = [b.id for b in self.branches]
        if len(set(branch_ids)) != len(branch_ids):
            raise AmbiguousBranchIDs
        
    @property
    def branch_ids(self) -> list[str]:
        return [b.id for b in self.branches if b.node1 in self.node_labels or b.node2 in self.node_labels]

    @property
    def node_labels(self) -> set[str]:
        connected_nodes = set([self.reference_node_label])
        nodes_to_assess = set([self.reference_node_label])
        nodes_already_assessed = set()
        while nodes_to_assess:
            new_nodes_to_assess = set()
            for node in nodes_to_assess:
                new_nodes_to_assess.update(self.nodes_connected_to(node))
            nodes_already_assessed.update(nodes_to_assess)
            nodes_to_assess = new_nodes_to_assess - nodes_already_assessed
            connected_nodes.update(new_nodes_to_assess)
        return connected_nodes

    @property
    def number_of_nodes(self) -> int:
        return len(self.node_labels)

    def is_zero_node(self, node: str) -> bool:
        return node == self.reference_node_label

    def branches_connected_to(self, node: str) -> list[Branch]:
        connected_branches = [branch for branch in self.branches if branch.node1 == node or branch.node2 == node]
        connected_branches.sort(key=lambda x: x.node1 if x.node1!=node else x.node2)
        return connected_branches

    def nodes_connected_to(self, node: str) -> set[str]:
        return {b.node1 if b.node1 != node else b.node2 for b in self.branches_connected_to(node=node)}

    def branches_between(self, node1: str, node2: str) -> list[Branch]:
        return [branch for branch in self.branches if set((branch.node1, branch.node2)) == set((node1, node2))]

    def __getitem__(self, id: str) -> Branch:
        if id not in [b.id for b in self.branches]:
            raise KeyError(f"Branch with id '{id}' not found in the network.")
        if id not in self.branch_ids:
            raise KeyError(f"Branch with id '{id}' is floating.")
        return {b.id: b for b in self.branches if b.id in self.branch_ids}[id]


import numpy as np
import itertools
from typing import Mapping
from ..network import Network
from . import matrix_operations as mo
from .matrix_operations import symbolic
from .. import transformers as trf
from .label_mapping import LabelMappingsFactory, default_label_mappings_factory
from .node_analysis_calculations import nodal_analysis_coefficient_matrix, nodal_analysis_constants_vector, source_incidence_matrix

def nodal_analysis_solution(network: Network, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), label_mappings_factory: LabelMappingsFactory = default_label_mappings_factory) -> tuple[complex | symbolic, ...]:
    label_mappings = label_mappings_factory(network)
    A = nodal_analysis_coefficient_matrix(network, matrix_ops=matrix_ops, label_mappings=label_mappings)
    b = nodal_analysis_constants_vector(network, matrix_ops=matrix_ops, label_mappings=label_mappings)
    try:
        return matrix_ops.solve(A, b)
    except mo.MatrixInversionException:
        return (float('nan'),) * len(b)

def open_circuit_impedance(network: Network, node1: str, node2: str, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), label_mappings_factory: LabelMappingsFactory = default_label_mappings_factory) -> complex | symbolic:
    if node1 == node2:
        return matrix_ops.elm(0).value
    if any([b.element.is_ideal_voltage_source for b in network.branches_between(node1, node2)]):
        return matrix_ops.elm(0).value
    if network.is_zero_node(node1):
        node1, node2 = node2, node1
    network = trf.switch_ground_node(network=network, new_ground=node2)
    label_mappings = label_mappings_factory(network)
    Y = nodal_analysis_coefficient_matrix(network, matrix_ops=matrix_ops, label_mappings=label_mappings)
    Y = matrix_ops.delete(Y, np.where(~matrix_ops.any_element(Y, axis=0))[0].tolist(), axis=1)
    Y = matrix_ops.delete(Y, np.where(~matrix_ops.any_element(Y, axis=1))[0].tolist(), axis=0)
    Z = matrix_ops.inv(Y)
    i1 = label_mappings.node_mapping[node1]
    return Z.diagonal()[i1]

def element_impedance(network: Network, element: str, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), label_mappings_factory: LabelMappingsFactory = default_label_mappings_factory) -> complex | symbolic:
    return open_circuit_impedance(
        network=trf.remove_element(network, element),
        node1=network[element].node1,
        node2=network[element].node2,
        matrix_ops=matrix_ops,
        label_mappings_factory=label_mappings_factory
    )

def state_space_matrices(network: Network, c_values: Mapping[str, float | symbolic] = {}, l_values: Mapping[str, float | symbolic] = {}, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), label_mappings_factory: LabelMappingsFactory = default_label_mappings_factory) -> tuple[mo.Matrix, mo.Matrix, mo.Matrix, mo.Matrix]:
    label_mappings = label_mappings_factory(network)
    def element_incidence_matrix(values: Mapping[str, float | symbolic]) -> np.ndarray:
        Delta = np.zeros((len(values), label_mappings.node_mapping.N))
        for (k, value), (i_label) in itertools.product(enumerate(values), label_mappings.node_mapping):
            if i_label == network[value].node1:
                Delta[k][label_mappings.node_mapping(i_label)] = +1
            if i_label == network[value].node2:
                Delta[k][label_mappings.node_mapping(i_label)] = -1
        return np.hstack((Delta, np.zeros((Delta.shape[0], label_mappings.voltage_source_mapper(network).N))))
    def source_and_inductance_incidence_matrix(values: Mapping[str, float | symbolic]) -> tuple[np.ndarray, np.ndarray]:
        Qi = source_incidence_matrix(network=network, label_mappings=label_mappings)
        Q = np.zeros((label_mappings.voltage_source_mapping.N, label_mappings.voltage_source_mapping.N), dtype=int)
        for i in label_mappings.voltage_source_mapping.values:
            Q[i][i] = 1
        Q = np.vstack((np.hstack( (Qi, np.zeros((Qi.shape[0], Q.shape[1]) ))),
                    np.hstack( (np.zeros((Q.shape[0], Qi.shape[1])), Q) )))
        QS = Q[:,[label_mappings.source_mapping[l] for l in label_mappings.source_mapping if l not in values]]
        QL = Q[:,[label_mappings.source_mapping[l] for l in label_mappings.source_mapping if l in values]]
        return QS, QL
    def value_matrix(c_values: Mapping[str, float | symbolic], l_values: Mapping[str, float | symbolic]) -> np.ndarray:
        return matrix_ops.vstack((
            matrix_ops.hstack(( matrix_ops.diag([-C for C in c_values.values()]), matrix_ops.zeros((len(c_values), len(l_values))) )), # type: ignore
            matrix_ops.hstack(( matrix_ops.zeros((len(l_values), len(c_values))), matrix_ops.diag([L for L in l_values.values()]) ))
        ))

    Delta = element_incidence_matrix(c_values)
    A_tilde = nodal_analysis_coefficient_matrix(network, matrix_ops=matrix_ops, label_mappings=label_mappings)
    QS, QL = source_and_inductance_incidence_matrix(l_values)
    DQ = np.hstack((Delta.T, QL))
    Lambda = value_matrix(c_values, l_values)
    invLambda = matrix_ops.diag([1/L for L in matrix_ops.diag_vec(Lambda)]) # type: ignore

    inv_A_tilde = matrix_ops.inv(A_tilde)
    transformed_inv_A_tilde = DQ.T @ inv_A_tilde
    sorted_A_tilde = matrix_ops.inv(transformed_inv_A_tilde @ DQ)

    A = invLambda @ sorted_A_tilde
    C = transformed_inv_A_tilde.T @ sorted_A_tilde
    B = (-invLambda @ C.T) @ QS
    D = (inv_A_tilde - transformed_inv_A_tilde.T @ C.T) @ QS

    return A, B, C, D



from .Components.components import Component
from .transformers import transformers
from .symbolic_transformers import transformers as symbolic_transformers
from ..Network.network import Network
import numpy as np
import sympy as sp
from dataclasses import dataclass, field

class AmbiguousComponentID(Exception): pass
class CircuitTransformationError(Exception): pass

@dataclass(frozen=True)
class Circuit:
    components : list[Component]
    ground_node : str | None = field(default=None)

    def __post_init__(self) -> None:
        if len(set([component.id for component in self.components])) != len(self.components):
            raise AmbiguousComponentID(f'Component list contains multiple components with the same ID.')

    def __getitem__(self, key: str) -> Component:
        index = [component.id for component in self.components].index(key)
        return self.components[index]

    def __iter__(self):
        return (component for component in self.components if component.type != 'ground')

def w(f: float) -> float:
    return 2*np.pi*f

def transform_circuit(circuit: Circuit, w: float, w_resolution: float = 1e-3, rms: bool = True) -> Network:
    reference_node_label = str(circuit.ground_node)
    if not circuit.ground_node:
        reference_node_label = circuit.components[0].nodes[0] if circuit.components else '0'
    try:
        return Network(
            branches=[transformers[component.type](component, w, w_resolution, rms) for component in circuit],
            reference_node_label=reference_node_label
        )
    except (ValueError, KeyError) as e:
        raise CircuitTransformationError from e

def transform(circuit: Circuit, w: list[float] = [0], w_resolution: float = 1e-3, rms: bool = True) -> list[Network]:
    return [transform_circuit(circuit, w_, w_resolution, rms) for w_ in w]

def frequency_components(circuit: Circuit, w_max: float) -> list[float]:
    def frequencies(component: Component) -> list[float]:
        try:
            w = float(component.value['w'])
        except KeyError:
            return []
        if component.type == 'periodic_voltage_source' or component.type == 'periodic_current_source':
            n_max = np.floor(w_max/w)
            return [w*n for n in np.arange(n_max+1)]
        return [w]
    return sorted(list(set([w for c in circuit.components for w in frequencies(c)])))

def transform_symbolic_circuit(circuit: Circuit, s: sp.Symbol = sp.Symbol('s', complex=True)) -> Network:
    reference_node_label = str(circuit.ground_node)
    if not circuit.ground_node:
        reference_node_label = circuit.components[0].nodes[0] if circuit.components else '0'
    try:
        return Network(
            branches=[symbolic_transformers[component.type](component, s) for component in circuit],
            reference_node_label=reference_node_label
        )   
    except (ValueError, KeyError) as e:
        raise CircuitTransformationError from e


from typing import Any, Callable
from .circuit import Circuit, Component
from .Components import components as cp
from .Components import symbolic_components as s_cp
from dataclasses import asdict
from .. import dump_load
import functools

class UnidentifiedComponent(Exception):
    ...

class UnknownCircuitComponent(Exception):
    ...

class IncorrectComponentInformation(Exception):
    ...

def numeric_component_factory(*_, factory_fcn: tuple[Callable[..., cp.Component], Callable[..., cp.Component]], numeric_keys: dict[str, type], **kwargs) -> cp.Component:
    def is_numeric(value: str) -> bool:
        try:
            complex(value)
            return True
        except ValueError:
            return False
    def probe_conversion(value: Any, target_type: type) -> Any:
        try:
            value = value.replace(' ', '').replace('\t', '')
        except AttributeError:
            pass
        try:
            return target_type(value)
        except (ValueError, TypeError):
            return value
    for key in numeric_keys.keys():
        kwargs[key] = probe_conversion(kwargs[key], numeric_keys[key])
    values = [kwargs[key] for key in numeric_keys]
    if all(is_numeric(value) for value in values):
        return factory_fcn[0](**kwargs)
    return factory_fcn[1](**kwargs)

circuit_component_translators : dict[str, Callable[..., cp.Component]] = {
    "resistor" : functools.partial(numeric_component_factory, factory_fcn=(cp.resistor, s_cp.resistor), numeric_keys={'R': float}),
    "conductance" : functools.partial(numeric_component_factory, factory_fcn=(cp.conductance, s_cp.conductance), numeric_keys={'G': float}),
    "impedance" : functools.partial(numeric_component_factory, factory_fcn=(cp.impedance, s_cp.impedance), numeric_keys={'Z': float}),
    "admittance" : functools.partial(numeric_component_factory, factory_fcn=(cp.admittance, s_cp.admittance), numeric_keys={'Y': float}),
    "dc_voltage_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.dc_voltage_source, s_cp.voltage_source), numeric_keys={'V': float}),
    "ac_voltage_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.ac_voltage_source, s_cp.voltage_source), numeric_keys={'V': float,  'f': float}),
    "open_circuit" : functools.partial(numeric_component_factory, factory_fcn=(cp.open_circuit, s_cp.open_circuit), numeric_keys={}),
    "short_circuit" : functools.partial(numeric_component_factory, factory_fcn=(cp.short_circuit, s_cp.short_circuit), numeric_keys={}),
    "complex_voltage_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.complex_voltage_source, s_cp.voltage_source), numeric_keys={'V': complex}),
    "dc_current_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.dc_current_source, s_cp.current_source), numeric_keys={'I': float}),
    "ac_current_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.ac_current_source, s_cp.current_source), numeric_keys={'I': float, 'f': float}),
    "complex_current_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.complex_current_source, s_cp.current_source), numeric_keys={'I': complex})
}

def generate_component(component: dict[str, Any]) -> Component:
    component = component.copy()
    try:
        component_id = component['id']
    except KeyError:
        raise UnidentifiedComponent(f'Unidentified component')
    try:
        component_value = component.pop('value', {})
    except KeyError:
        raise IncorrectComponentInformation(f"Missing value of component '{component_id}'.")
    try:
        component_type = component.pop('type')
    except KeyError:
        raise IncorrectComponentInformation(f"Missing type information of component '{component_id}'.")
    try:
        component_nodes = component['nodes']
    except KeyError:
        raise IncorrectComponentInformation(f"Missing node information of component '{component_id}'.")
    try:
        component_factory = circuit_component_translators[component_type]
    except KeyError:
        raise UnknownCircuitComponent(f"Unknown type '{component_type}' of component '{component_id}' is unknown.")
    try:
        return component_factory(id=component_id, nodes=component_nodes, **component_value)
    except KeyError as e:
        raise IncorrectComponentInformation(f"Missing information '{e.args[0]}' for component '{component_id}' of type '{component_type}'.") from e

def undictify_circuit(circuit: dict) -> Circuit:
    return Circuit([generate_component(entry) for entry in circuit['components']], ground_node=circuit.get('ground_node', ''))

deserialize = functools.partial(dump_load.deserialize, dict_preprocessor=undictify_circuit)
load = functools.partial(dump_load.load, deserialize_fcn=deserialize)

def dictify_circuit(circuit: Circuit) -> dict:
    return {'components' : [asdict(c) for c in circuit.components], 'ground_node' : circuit.ground_node}

serialize = functools.partial(dump_load.serialize, dict_processor=dictify_circuit)
save = functools.partial(dump_load.dump, dump_fcn=serialize)



from .network import Network, Branch
from .norten_thevenin_elements import NortenTheveninElement
from . import elements as elm
from . import symbolic_elements as selm
import sympy as sp

def switch_ground_node(network: Network, new_ground: str) -> Network:
    return Network(network.branches, new_ground)

def remove_element(network: Network, element: str) -> Network:
    branch_to_remove = network[element]
    if branch_to_remove.node1 == network.reference_node_label and len(network.branches_connected_to(branch_to_remove.node1)) == 1:
       network = rename_node(network, branch_to_remove.node2, network.reference_node_label) 
    if branch_to_remove.node2 == network.reference_node_label and len(network.branches_connected_to(branch_to_remove.node2)) == 1:
       network = rename_node(network, branch_to_remove.node1, network.reference_node_label) 
    branches = [b for b in network.branches if b.element.name != element]
    return Network(branches, reference_node_label=network.reference_node_label)

def rename_node(network: Network, old_node: str, new_node: str) -> Network:
    def rename(branch: Branch) -> Branch:
        if branch.node1 == old_node:
            branch = Branch(new_node, branch.node2, branch.element)
        if branch.node2 == old_node:
            branch = Branch(branch.node1, new_node, branch.element)
        return branch
    branches = [rename(b) for b in network.branches]
    return Network(branches, reference_node_label=network.reference_node_label)

def remove_active_elements(network: Network) -> Network:
    def impedance(e: NortenTheveninElement) -> NortenTheveninElement:
        try:
            Z = complex(e.Z)
        except TypeError:
            return selm.impedance(name=e.name, Z=sp.sympify(e.Z))
        return elm.impedance(name=e.name, Z=Z)
    def admittance(e: NortenTheveninElement) -> NortenTheveninElement:
        try:
            Y = complex(e.Y)
        except TypeError:
            return selm.admittance(name=e.name, Y=sp.sympify(e.Y))
        return elm.admittance(name=e.name, Y=Y)
    def remove(b: Branch) -> Branch:
        if b.element.is_ideal_voltage_source:
            return Branch(b.node1, b.node2, elm.open_circuit(b.element.name))
        if b.element.is_ideal_current_source:
            return Branch(b.node1, b.node2, elm.open_circuit(b.element.name))
        if b.element.is_voltage_source:
            return Branch(b.node1, b.node2, impedance(b.element))
        if b.element.is_current_source:
            return Branch(b.node1, b.node2, admittance(b.element))
        return b
    return Network(
        branches=[remove(b) for b in network.branches],
        reference_node_label=network.reference_node_label
    )